(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory() :
  typeof define === 'function' && define.amd ? define(factory) :
  (global.Topologica = factory());
}(this, (function () { 'use strict';

  function Graph () {
    var edges = {};

    var adjacent = function (node) { return edges[node] || []; };

    var addNode = function (node) {
      edges[node] = adjacent(node);
    };

    var addEdge = function (u, v) {
      addNode(u);
      addNode(v);
      adjacent(u).push(v);
    };

    var depthFirstSearch = function (sourceNodes) {
      var visited = new Set();
      var nodeList = [];

      function DFSVisit(node) {
        if (!visited.has(node)) {
          visit(node);
          nodeList.push(node);
        }
      }
      function visit(node) {
        visited.add(node);
        adjacent(node).forEach(DFSVisit);
      }

      sourceNodes.forEach(visit);

      return nodeList;
    };

    var topologicalSort = function (sourceNodes) { return depthFirstSearch(sourceNodes)
        .reverse(); };

    return {
      addEdge: addEdge,
      topologicalSort: topologicalSort
    };
  }

  var isAsync = function (fn) { return fn.constructor.name === 'AsyncFunction'; };

  var parse = function (dependencies) { return dependencies.split
    ? dependencies.split(',').map(function (str) { return str.trim(); })
    : dependencies; };

  var Topologica = function (options) {
    var values = new Map();
    var functions = new Map();
    var graph = Graph();
    var changed = new Set();

    var invoke = function (property) {
      functions.get(property)();
    };

    var digest = function () {
      graph
        .topologicalSort(Array.from(changed.values()))
        .forEach(invoke);
      changed.clear();
    };

    var set = function (options) {
      Object.keys(options).forEach(function (property) {
        var value = options[property];
        if (values.get(property) !== value) {
          values.set(property, value);
          changed.add(property);
        }
      });
      digest();
    };

    var get = values.get.bind(values);

    var getAll = function (properties) { return properties.reduce(function (accumulator, property) {
        accumulator[property] = get(property);
        return accumulator;
      }, {}); };

    var allDefined = function (properties) { return properties
      .every(values.has, values); };

    if (options) {
      Object.keys(options).forEach(function (property) {
        var fn = options[property];
        var dependencies = parse(fn.dependencies);

        var propertySync = isAsync(fn) ? property + "'" : property;

        dependencies.forEach(function (input) {
          graph.addEdge(input, propertySync);
        });

        functions.set(propertySync, function () {
          if (allDefined(dependencies)) {
            var output = fn(getAll(dependencies));
            isAsync(fn)
              ? output.then(function (value) {
                var obj;

                set(( obj = {}, obj[property] = value, obj ));
              })
              : values.set(property, output);
          }
        });
      });
    }

    return { set: set, get: get };
  };

  return Topologica;

})));
